//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtkmm.h>
#include <glibmm.h>
#include <glibmm/i18n.h>
#include <glib/gstdio.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <unistd.h>
#include <sys/stat.h>

#include "util.hpp"
#include "uri++.hpp"
#include "file_utils.hpp"
#include "vfs.hpp"
#include "network.hpp"
#include "debug.hpp"

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xmlreader.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>

#include "xml.hpp"
#include "main.hpp"
#include "paths.hpp"
#include "musicbrainz_ws.hpp"
#include "amazon.hpp"

#include "x_library.hpp"
#include "x_vfs.hpp"

#define AMAZON_FORMAT_STRING_URI_ASIN "http://images.amazon.com/images/P/%s.01.MZZZZZZZ.jpg"

namespace Bmp
{
  namespace Amazon
  {
   void 
    get_cover_by_asin (const Glib::ustring& asin, Glib::RefPtr<Gdk::Pixbuf>& cover, bool only_cached)
    {
	using namespace std;

	using namespace Bmp;
	using namespace Glib;
	using namespace Gdk;

	if (!Network::is_connected ())
	  {
	    throw NO_COVER;
	  }

	string basename = (boost::format ("%s.png") % asin).str ();
	string thumb_path = Glib::build_filename (BMP_PATH_COVER_THUMB_DIR, basename);
	if (Glib::file_test (thumb_path, Glib::FILE_TEST_EXISTS))
	  {
	      cover = Gdk::Pixbuf::create_from_file (thumb_path); 
	      return;
	  }

	if (only_cached)
	  {
	      throw NO_COVER;
	  }

	URI uri ((boost::format (AMAZON_FORMAT_STRING_URI_ASIN) % asin).str());
	uri.escape ();
	VFS::Handle handle = VFS::Handle::create (ustring(uri));
	vfs->read (handle, VFS::TRANSPORT);
    
	if (!handle.get_buffer())
	{
	  throw NO_COVER;
	}

	RefPtr<Gdk::PixbufLoader> loader = PixbufLoader::create ("jpeg"); 
	loader->write (reinterpret_cast<const guint8*>(handle.get_buffer()), handle.get_buffer_size());
	loader->close ();
	cover = loader->get_pixbuf()->scale_simple ( 128, 128, INTERP_HYPER );
    }

    void 
    get_cover_artist_album (const Glib::ustring& artist,
			    const Glib::ustring& album,
			    Glib::RefPtr<Gdk::Pixbuf>& cover,
			    Glib::ustring& asin,
			    bool only_cached) 
    {
	using namespace Glib;
	using namespace Bmp::Library;
	using namespace std;

	MusicBrainzWS::MB_Releases mb_releases;
	MusicBrainzWS::get_releases (artist, album, mb_releases);
	if (mb_releases.empty())
	  {
	    throw NO_COVER;	
	  }

	MusicBrainzWS::MB_Release release = mb_releases[0];
	try {
	  get_cover_by_asin (release.asin, cover, only_cached);
	  string basename = (boost::format ("%s.png") % release.asin).str ();
	  string thumb_path = Glib::build_filename (BMP_PATH_COVER_THUMB_DIR, basename);
	  cover->save (thumb_path, "png");
	  asin = release.asin;
	} catch (...) {}
    }

    void 
    get_cover (const Glib::ustring& _uri, Glib::RefPtr<Gdk::Pixbuf>& cover, bool only_cached) 
    {
	using namespace std;
	using namespace Glib;
	using namespace Bmp::Library;

	debug ("amazon", "Here now");

	if (!Network::is_connected ())
	  {
	    debug ("amazon", "No Network connection");
	    throw NO_COVER;
	  }

	Bmp::DB::DataRow row;
	library->get_metadata (_uri, row); 

	Track track (row);
	if (track.asin)
	  {
	    debug ("amazon", "Getting cover by asin: %s", track.asin.get().c_str());
	    get_cover_by_asin (track.asin.get(), cover);
	    return;
	  }

	if (!(track.artist && track.album))
	  {
	    debug ("amazon", "No artist & album");
	    throw NO_COVER;
	  }

	MusicBrainzWS::MB_Releases mb_releases;
	MusicBrainzWS::get_releases (track.artist.get(), track.album.get(), mb_releases);
	if (mb_releases.empty())
	  {
	    debug ("amazon", "No releases.");
	    throw NO_COVER;	
	  }

	MusicBrainzWS::MB_Release release = mb_releases[0];
	string basename = (boost::format ("%s.png") % release.asin).str ();
	string thumb_path = Glib::build_filename (BMP_PATH_COVER_THUMB_DIR, basename);

	debug ("amazon", "release.asin: %s", release.asin.c_str());

	get_cover_by_asin (release.asin, cover, only_cached);
	cover->save (thumb_path, "png");
	library->set_asin_for_album (track.artist.get(), track.album.get(), release.asin);
    }
  }
}
