#include <glibmm/i18n.h>
#include <glib/gstdio.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>
#include <fstream>

#include <boost/format.hpp>

#include "file_utils.hpp"
#include "ui_util.hpp"
#include "util.hpp"
#include "paths.hpp"
#include "main.hpp"
#include "debug.hpp"
#include "shell.hpp"
#include "xds.hpp"
#include "uri++.hpp"

#include "library_ui.hpp"
#include "library_ui_merge.hpp"
#include "musicbrainz_ws.hpp"

#include "x_library.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

namespace
{
    using namespace Bmp::Library;

    const Datum search_data[] =
    {
	DATUM_ARTIST,
	DATUM_ALBUM,
	DATUM_TITLE,
	DATUM_GENRE,
    };

    const int search_default[] =
    {
	0, 1, 2,
    };
}

namespace Bmp
{
    class TreeViewXDS : public Gtk::TreeView, public Bmp::XDS
    {
	public:
  
	  TreeViewXDS () {}
	  ~TreeViewXDS () {}

	  virtual void
	  on_drag_end (const Glib::RefPtr<Gdk::DragContext>& context)
	  {
	    xds_drag_end (context);
	  }

	  virtual void
	  on_drag_begin  (const Glib::RefPtr<Gdk::DragContext>& context)
	  {
	    Glib::ustring dir_basename;
	    dir_basename.append (artist.get());
	    dir_basename.append (" - ");
	    dir_basename.append (album.get());
	    dir_basename.append (".mlq");

	    context->set_icon (pixbuf, 0, 0);	    

	    xds_drag_begin (context, dir_basename); 
	  }

	  virtual void
	  on_drag_data_get (const Glib::RefPtr<Gdk::DragContext>& context,
			    Gtk::SelectionData& selection_data, guint info, guint time)
	  {
	    xds_drag_data_get (context, selection_data, info, time);
	  }

	  void
	  set_album  (const Glib::ustring& album)
	  {
	    this->album  = album;
	  }

	  void
	  set_artist (const Glib::ustring& artist)
	  {
	    this->artist = artist;
	  }

	  void
	  clear_artist_album ()
	  {
	    artist.reset ();
	    album.reset ();
	  }
    
	  void
	  set_pixbuf (Glib::RefPtr<Gdk::Pixbuf>& pixbuf)
	  {
	    this->pixbuf = pixbuf;
	  }

	protected:

	  virtual void
	  xds_drag_action     (const Glib::ustring& uri)
	  {
	      using namespace Bmp::DB;
	      using namespace Bmp::Library;

	      std::string id_artist	= library->get_metadatum_id (DATUM_ARTIST);
	      std::string id_album	= library->get_metadatum_id (DATUM_ALBUM);

	      dynamic_cast<Gtk::Widget *>(this)->set_sensitive (false);

	      Glib::ustring outstr;

	      outstr.append ("mlq:///?");

	      outstr.append (id_artist);
	      outstr.append ("=");
	      outstr.append (artist.get());

	      outstr.append ("&");
	  
	      outstr.append (id_album);
	      outstr.append ("=");
	      outstr.append (album.get());

	      Bmp::URI u (outstr);
	      u.escape ();

	      std::ofstream outfile;
	      outfile.open (Glib::filename_from_uri (uri).c_str());
	      outfile << "#MLQ\n"; 
	      outfile << Glib::ustring (u).c_str(); 
	      outfile << "\n";
	      outfile.close ();

	      dynamic_cast<Gtk::Widget *>(this)->set_sensitive (true);
	  }
	
	  virtual bool
	  xds_acceptable      (const Glib::ustring& uri)
	  {
	      if (!artist && !album) return false;
	      return true;
	  }

	private:
    
	  boost::optional<Glib::ustring> artist;
	  boost::optional<Glib::ustring> album;
	  Glib::RefPtr<Gdk::Pixbuf> pixbuf;
    };
}

namespace Bmp
{
      void
      LibraryUI::notebook_page_switch (GtkNotebookPage *page, unsigned int id)
      {
	  switch (LibraryPage(id))
	  {
	      case PAGE_ALBUMS: update_track_view_from_albums (); break;
	      case PAGE_SEARCH: update_track_view_from_search (); break;
	  }
      }

      LibraryUI::LibraryUI (BaseObjectType                        *cobject,
			    const Glib::RefPtr<Gnome::Glade::Xml> &xml)
	: Gtk::Window (cobject),
	  ref_xml (xml),
	  updated (false),
	  busy (false)
      {
	Gtk::Image *image;

	bmp_window_set_icon_list (GTK_WIDGET (gobj ()), "library");

	ref_xml->get_widget ("notebook_library", notebook_library);
	conn_notebook = notebook_library->signal_switch_page().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::notebook_page_switch));

	ref_xml->get_widget ("i_header", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "library" G_DIR_SEPARATOR_S "header-logo.png"));

	ref_xml->get_widget ("i_add", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "library" G_DIR_SEPARATOR_S "add.png"));
	ref_xml->get_widget ("i_wait1", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "wait.gif"));
	ref_xml->get_widget ("i_search", image);
	image->set (Gtk::StockID (BMP_STOCK_SEARCH), Gtk::ICON_SIZE_BUTTON);


	ref_xml->get_widget ("i_remove_album", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "library" G_DIR_SEPARATOR_S "remove.png"));
	ref_xml->get_widget ("i_remove_tracks", image);
	image->set (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "library" G_DIR_SEPARATOR_S "remove.png"));

	ref_xml->get_widget ("i_musicbrainz", image);
	image->set (Gtk::StockID (GTK_STOCK_CONNECT), Gtk::ICON_SIZE_SMALL_TOOLBAR);


	ref_xml->get_widget ("b_add",
			      b_add);
	b_add->signal_clicked().connect			(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::music_add));

	ref_xml->get_widget ("b_remove_tracks",
			      b_remove_tracks);
	b_remove_tracks->signal_clicked().connect       (sigc::mem_fun (*this,
							      &Bmp::LibraryUI::remove_tracks));

	ref_xml->get_widget ("b_remove_album",
			      b_remove_album);
	b_remove_album->signal_clicked().connect       (sigc::mem_fun (*this,
							      &Bmp::LibraryUI::remove_album));

	ref_xml->get_widget ("b_close",
			      b_close);
	b_close->signal_clicked().connect		(sigc::mem_fun (*this,
							      &Gtk::Widget::hide));

	ref_xml->get_widget ("b_play",
			      b_play); 
	b_play->signal_clicked().connect		(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::play));

	ref_xml->get_widget ("b_enqueue",
			      b_enqueue);
	b_enqueue->signal_clicked().connect		(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::enqueue));
	ref_xml->get_widget ("b_select_all",
			      b_select_all);
	b_select_all->signal_clicked().connect		(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::select_all));

	ref_xml->get_widget ("b_albums_merge",
			      b_albums_merge);
	b_albums_merge->signal_clicked().connect	(sigc::mem_fun (*this,
							      &Bmp::LibraryUI::albums_merge));

	// Search widgets
	using namespace Bmp::Library;
	using namespace Gtk;

	cs_store = Gtk::ListStore::create (cr_search);

	for (unsigned int n = 0; n < G_N_ELEMENTS (search_data); ++n) 
	{
	  TreeModel::iterator i = cs_store->append ();
	  DatumDefine define = library->get_metadatum_define (search_data[n]);
	  (*i)[cr_search.name] = define.title; 
	  (*i)[cr_search.id] = define.id;
	}

	static boost::format f_cbox     ("cbox_%d");
	static boost::format f_cbutton  ("cb_%d");
	static boost::format f_entry    ("entry_%d");
	for (int n = 0; n < N_FILTERS; ++n)
	{
	    ref_xml->get_widget ((f_cbutton % (n+1)).str(), cbutton[n]);
	    ref_xml->get_widget ((f_cbox    % (n+1)).str(), cbox[n]);
	    ref_xml->get_widget ((f_entry   % (n+1)).str(), entry[n]);

	    cbutton[n]->signal_toggled ().connect (sigc::bind (sigc::mem_fun (*this, &Bmp::LibraryUI::filter_toggled), FilterID(n)));
	    cbox[n]->clear ();
	    Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
	    cbox[n]->pack_start (*cell);
	    cbox[n]->add_attribute (*cell, "text", 0);
	    cbox[n]->set_model (cs_store);
	    cbox[n]->set_active  (search_default[n]);

	    entry[n]->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::update_track_view_from_search));
	    cbox[n]->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::update_track_view_from_search));
	    cbutton[n]->signal_toggled().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::update_track_view_from_search));
	}

	// Watch shutdown-request from SystemControl
	g_signal_connect (G_OBJECT (bmp_system_control), "shutdown-request", G_CALLBACK (Bmp::LibraryUI::shutdown_request), this);

	ref_xml->get_widget ("notebook_add",  notebook_add);
	ref_xml->get_widget ("view_tracks",   view_tracks);
	ref_xml->get_widget ("albums_filter", albums_filter);

	view_albums = new Bmp::TreeViewXDS;
	view_albums->show_all ();
	view_albums->set_headers_visible (false);

	Gtk::ScrolledWindow *sw;
	ref_xml->get_widget ("sw_albums", sw);
	sw->add (*view_albums);

	//Tracks view
	struct { 
	    char    *title;
	    double   xalign;
	    int	     width;
	} headers_tracks[] = {
	    { _("Track"),      1.0, 30  },
	    { _("Title"),      0.0, 100 },
	    { _("Artist"),     0.0, 80  },
	    { _("Album"),      0.0, 80  },
	    { _("Genre"),      0.0, 80  },
	    { _("Bitrate"),    0.0, 30  },
	    { _("Samplerate"), 0.0, 30  },
	};

	for (unsigned int n = 0; n < G_N_ELEMENTS(headers_tracks); ++n)
	{
	  Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );

	  cell->property_ellipsize () = Pango::EllipsizeMode(PANGO_ELLIPSIZE_END);
	  cell->property_xalign() = headers_tracks[n].xalign;

	  view_tracks->append_column (headers_tracks[n].title, *cell);
	  Gtk::TreeViewColumn *column = view_tracks->get_column (n);

	  column->add_attribute (*cell, "text", n);
	  column->set_resizable (true);
	  column->set_sort_column (n);
	  column->set_expand (true);
	  column->property_min_width() = headers_tracks[n].width;
	}

	store_tracks = Gtk::ListStore::create (tracks);

	view_tracks->set_model (store_tracks);
	view_tracks->get_selection()->set_mode (Gtk::SELECTION_MULTIPLE);
	view_tracks->get_selection()->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::tracks_selection_changed));


	//Albums view
	enum Renderer
	{
	  R_TEXT,
	  R_PIXBUF,
	};

	struct { 
	    char    *title;
	    double   xalign;
	    int	     width;
	    Renderer renderer;
	} headers_albums[] = {
	    { _("Cover"),      0.5, 72,  R_PIXBUF },
	    { _("Album"),      0.0, 250, R_TEXT   },
	};

	for (unsigned int n = 0; n < G_N_ELEMENTS(headers_albums); ++n)
	{
	  switch (headers_albums[n].renderer)
	  {
		case R_PIXBUF:
		{
		  Gtk::CellRendererPixbuf *cell = Gtk::manage ( new Gtk::CellRendererPixbuf() );

		  cell->property_xalign() = headers_albums[n].xalign;
		  cell->property_xpad() = 3;
		  cell->property_ypad() = 4;

	  	  view_albums->append_column (headers_albums[n].title, *cell);
		  Gtk::TreeViewColumn *column = view_albums->get_column (n);

	  	  column->add_attribute (*cell, "pixbuf", n);
	  	  column->set_resizable (false);
	  	  column->set_expand (false);
	  	  column->set_sizing (Gtk::TREE_VIEW_COLUMN_FIXED);
		  column->property_min_width() = headers_albums[n].width;
		  column->property_max_width() = headers_albums[n].width;

		  break;
		}

		case R_TEXT:
		{
		  Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );

		  cell->property_ellipsize () = Pango::ELLIPSIZE_END;
		  cell->property_xalign() = headers_albums[n].xalign;
		  cell->property_size() = 14 * PANGO_SCALE;
		  cell->property_width () = 700;

	  	  view_albums->append_column (headers_albums[n].title, *cell);
		  Gtk::TreeViewColumn *column  = view_albums->get_column (n);

	  	  column->add_attribute (*cell, "markup", n);
	  	  column->set_resizable (false);
	  	  column->set_expand (false);
	  	  column->set_sizing (Gtk::TREE_VIEW_COLUMN_FIXED);
		  column->property_min_width() = headers_albums[n].width;
		  column->property_max_width() = headers_albums[n].width;

		  break;
		}
	  }

	}

	store_albums = Gtk::ListStore::create (albums);
	store_albums->set_default_sort_func (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_default_sort_func));
	store_albums->set_sort_column_id (-1, Gtk::SORT_ASCENDING);

	store_albums_filter = Gtk::TreeModelFilter::create (store_albums);
	store_albums_filter->set_visible_func (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_row_is_visible));

	view_albums->set_model (store_albums_filter);
	view_albums->set_fixed_height_mode ();
	view_albums->get_selection()->set_mode (Gtk::SELECTION_MULTIPLE);
	view_albums->get_selection()->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::update_track_view_from_albums));
	view_albums->signal_event ().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_event));
	view_albums->signal_row_activated ().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::play_current));

	albums_filter->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_refilter));

	//Connect to the library for blocking the UI to avoid race conditions
	library->signal_processing_start().connect (sigc::bind (sigc::mem_fun (*this, &Gtk::Widget::set_sensitive), false));
	library->signal_processing_end().connect (sigc::bind (sigc::mem_fun (*this, &Gtk::Widget::set_sensitive), true));
      }
      
      LibraryUI::~LibraryUI ()
      {
	conn_notebook.disconnect ();
	delete view_albums; 
	ref_xml.clear();
      }

      LibraryUI*
      LibraryUI::create ()
      {
	const std::string path = Glib::build_filename (DATA_DIR, "glade" G_DIR_SEPARATOR_S "library_ui.glade");
	Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
	LibraryUI *library_ui = 0;
	glade_xml->get_widget_derived ("library_ui", library_ui);
	return library_ui;
      }

      //////////////////////////// Search 
      void
      LibraryUI::filter_toggled (FilterID id)
      {
	bool active = cbutton[int(id)]->get_active();
	cbox[int(id)]->set_sensitive (active);
	entry[int(id)]->set_sensitive (active);
      }

      //////////////////////////// Handling of covers/cover-fetching
      void
      LibraryUI::fetch_cover (const Gtk::TreePath& path)
      { 
	using namespace Gtk;
	using namespace Gdk;
	using namespace Glib;
	using namespace std;

	TreeModel::iterator i_Selected = store_albums_filter->convert_iter_to_child_iter (store_albums_filter->get_iter (path));
	if ((*i_Selected)[albums.has_cover]) return;

	this->set_sensitive (false); //FIXME: Don't set the entire window insensitive
	RefPtr<Pixbuf> cover;
	try {

		Glib::ustring asin = library->get_asin_for_album ((*i_Selected)[albums.album]);

		try { 
		    Amazon::get_cover_by_asin (asin, cover, false);

		    if (cover)
		    {
		      library->set_asin_for_album ((*i_Selected)[albums.artist], (*i_Selected)[albums.album], asin);

		      (*i_Selected)[albums.cover] = cover->scale_simple (64, 64, INTERP_BILINEAR);
		      (*i_Selected)[albums.has_cover] = true;
		    }
		} catch (...) {}

	} catch (Bmp::Library::Exception& cxe) {

		Glib::ustring asin;

		try {
		    Amazon::get_cover_artist_album ((*i_Selected)[albums.artist], (*i_Selected)[albums.album], cover, asin, false);

		    if (cover)
		    {
		      library->set_asin_for_album ((*i_Selected)[albums.artist], (*i_Selected)[albums.album], asin);

		      (*i_Selected)[albums.cover] = cover->scale_simple (64, 64, INTERP_BILINEAR);
		      (*i_Selected)[albums.has_cover] = true;
		    }
		} catch (...) {}

	} catch (...) {}
	this->set_sensitive (true);
      }

      bool
      LibraryUI::albums_event (GdkEvent * ev)
      {
	GdkEventButton *event = 0;
	if (ev->type == GDK_BUTTON_PRESS)
	{
	  event = reinterpret_cast<GdkEventButton *>(ev);
	  Gtk::TreeViewColumn *column, *column_0 = view_albums->get_column (0);
	  Gtk::TreeModel::Path path;
	  int cell_x, cell_y;

	  if (view_albums->get_path_at_pos (int(event->x),
					    int(event->y),
					    path,
					    column,
					    cell_x,
					    cell_y)) 
	  {
	    if (column == column_0)
	    {
		if (event->type == GDK_BUTTON_PRESS)
		  {
		    fetch_cover (path);
		    return true;
		  }
	    }
	  }
	} 
	return false;
      }

      //////////////////////////// Sorting
      int 
      LibraryUI::albums_default_sort_func (const Gtk::TreeModel::iterator& iter_a, const Gtk::TreeModel::iterator& iter_b)
      {
	Glib::ustring artist_a = (*iter_a)[albums.artist]; 
	Glib::ustring artist_b = (*iter_b)[albums.artist]; 
	Glib::ustring album_a  = (*iter_a)[albums.album];
	Glib::ustring album_b  = (*iter_b)[albums.album];

	Glib::ustring str_a = artist_a + " " + album_a; 
	Glib::ustring str_b = artist_b + " " + album_b; 

	return str_a.compare(str_b);
      }

      //////////////////////////// Selections
      void
      LibraryUI::tracks_selection_changed ()
      {
	  bool selected = view_tracks->get_selection ()->count_selected_rows ();

	  b_play-> set_sensitive (selected);
	  b_enqueue-> set_sensitive (selected);
	  b_remove_tracks-> set_sensitive (selected);
      }

      void
      LibraryUI::fill_tracks_from_rows (const Bmp::DB::VDataRow& vector) 
      {
	  using namespace Bmp::DB;
	  using namespace Bmp::Library;
	  using namespace Gtk;

	  for (VDataRow::const_iterator v = vector.begin (), end = vector.end (); v != end; ++v) 
	  { 
	    TreeModel::iterator tracks_iter = store_tracks->append ();
	    Track track (*v);

	    if (track.title) (*tracks_iter)[tracks.title] = track.title.get (); 
	    if (track.artist) (*tracks_iter)[tracks.artist] = track.artist.get (); 
	    if (track.album) (*tracks_iter)[tracks.album] = track.album.get (); 
	    if (track.tracknumber) (*tracks_iter)[tracks.track] = track.tracknumber.get (); 
	    if (track.genre) (*tracks_iter)[tracks.genre] = track.genre.get (); 
	    if (track.bitrate) (*tracks_iter)[tracks.bitrate] = track.bitrate.get (); 
	    if (track.samplerate) (*tracks_iter)[tracks.samplerate] = track.samplerate.get (); 
	    if (track.location) (*tracks_iter)[tracks.uri] = track.location.get (); 
	  }
      }

      // Search
      void
      LibraryUI::update_track_view_from_search ()
      {
	  using namespace Bmp::DB;
	  using namespace Bmp::Library;
	  using namespace Gtk;

	  busy = true;

          store_tracks->clear ();
	  b_remove_album->set_sensitive (false);
	  b_albums_merge->set_sensitive (false);

	  AttributeList attrs; 
	  VDataRow vector;

	  for (unsigned int n = 0; n < 3; ++n)
	  {
	      if (cbutton[n]->get_active ())
	      {
		Datum datum = search_data[cbox[n]->get_active_row_number()]; 
		std::string id = library->get_metadatum_id (datum);
		if (!entry[n]->get_text().empty())
		{
		  attrs.push_back (Attribute (false, id, entry[n]->get_text())); 
		}
	      }
	  }

	  if (attrs.empty())
	  {
	      busy = false;
	      return;
	  }

	  library->query (attrs, vector);
	  fill_tracks_from_rows (vector);

	  if (!store_tracks->children ().empty ())
	  {
	    b_select_all->set_sensitive (true);
	  }

	  busy = false;
      }

      // Albums	
      void
      LibraryUI::update_track_view_from_albums ()
      {
	  using namespace Bmp::DB;
	  using namespace Bmp::Library;
	  using namespace Gtk;

	  if (busy) return;

	  busy = true;

          store_tracks->clear ();
	  b_remove_album->set_sensitive (false);

	  if (!view_albums->get_selection ()->count_selected_rows ())
	  {
	    debug ("library-ui", "No Albums selected.");
	    busy = false;
	    return;
	  }

          TreeSelection::ListHandle_Path paths = view_albums->get_selection()->get_selected_rows ();

	  if (paths.size () == 1)
	  {
	    //FIXME: We should possibly allow for removal of several albums in a run
	    b_remove_album->set_sensitive (true);
#ifdef USE_AMAZON
	    //FIXME: This is a little awkward as we do an iteration below anyway
	    //FIXME: We should possibly allow for fetching of several albums in a run
	    TreeModel::iterator cur = store_albums_filter->get_iter (*(paths.begin())); 

	    if ((*cur)[albums.has_cover])
	      {
		Glib::RefPtr<Gdk::Pixbuf> cover = (*cur)[albums.cover]; 
		view_albums->set_pixbuf (cover);
	      }
	    else
	      {
		Glib::RefPtr<Gdk::Pixbuf> disc = 
		      Gdk::Pixbuf::create_from_file (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "audio.png"))->scale_simple 
					  (64, 64, Gdk::INTERP_BILINEAR);
		view_albums->set_pixbuf (disc);
	      }

	    view_albums->set_artist ((*cur)[albums.artist]);
	    view_albums->set_album ((*cur)[albums.album]);
	    view_albums->xds_set ();
#endif
	  }
	  else
	  {
	      view_albums->clear_artist_album ();
	      view_albums->xds_unset ();
	  }

	  b_albums_merge->set_sensitive (true);

	  std::string id_artist = library->get_metadatum_id (DATUM_ARTIST);
	  std::string id_album  = library->get_metadatum_id (DATUM_ALBUM);

	  debug ("library-ui", "Updating Tracks View from Albums.");
          for (TreeSelection::ListHandle_Path::const_iterator path = paths.begin (); path != paths.end (); ++path)
          {
	    TreeModel::iterator cur = store_albums_filter->get_iter (*path);

	    Glib::ustring key_A = (*cur)[albums.artist];
	    Glib::ustring key_B = (*cur)[albums.album];

	    AttributeList attrs; 
	    VDataRow vector;

	    attrs.push_back (Attribute (true, id_artist, key_A));
	    attrs.push_back (Attribute (true, id_album,  key_B));

	    library->query (attrs, vector);
	    fill_tracks_from_rows (vector);
	  }

	  if (!store_tracks->children ().empty ())
	  {
	    b_select_all->set_sensitive (true);
	  }

	  busy = false;
      }

      void
      LibraryUI::select_all ()
      {
	view_tracks->get_selection()->select_all ();
      }

      //////////////////////////// Removing
      void
      LibraryUI::remove_album ()
      {
	  using namespace Bmp::DB;
	  using namespace Bmp::Library;
	  using namespace Glib;
	  using namespace Gtk;
	  using namespace std;

	  TreeSelection::ListHandle_Path paths = view_albums->get_selection()->get_selected_rows ();
	  TreeModel::iterator selected = store_albums_filter->convert_iter_to_child_iter (store_albums_filter->get_iter (*(paths.begin()))); 

          AttributeList attrs;

          ustring key_A = (*selected)[albums.artist];
          ustring key_B = (*selected)[albums.album];

          attrs.push_back (Attribute (true, library->get_metadatum_id (DATUM_ARTIST), string(key_A)));
          attrs.push_back (Attribute (true, library->get_metadatum_id (DATUM_ALBUM),  string(key_B)));

          library->del (attrs);
          store_albums->erase (selected);
      }

      void
      LibraryUI::remove_tracks ()
      {
	  using namespace Bmp::DB;
	  using namespace Bmp::Library;
	  using namespace Glib;
	  using namespace Gtk;
	  using namespace std;

	  typedef std::vector<Gtk::TreeModel::iterator> IterList;
	  IterList iters;

	  bool physically_delete_files = mcs->key_get<bool>("bmp", "physically-delete-files");
	  if (physically_delete_files)
	  {
	    Gtk::MessageDialog dialog (*this, "<b>You are about to delete the selected files also <i>physically</i> from the filesystem.\nIf you do NOT wish that, please press cancel now.</b>\n\n<small>NOTE: You can change this option in Preferences, section 'Library'.</small>", true, Gtk::MESSAGE_WARNING, Gtk::BUTTONS_OK_CANCEL, true);
	    if (dialog.run() != GTK_RESPONSE_OK)
	    {
	      dialog.hide ();
	      return;
	    }
	  }

          TreeSelection::ListHandle_Path paths = view_tracks->get_selection()->get_selected_rows ();
          for (TreeSelection::ListHandle_Path::const_iterator iter = paths.begin (); iter != paths.end (); ++iter)
	  {
	      iters.push_back (store_tracks->get_iter (*iter));
	  }

          for (IterList::const_iterator cur = iters.begin (), end = iters.end (); cur != end; ++cur)
          {
              ustring key = (*(*cur))[tracks.uri];
              AttributeList attrs;
              attrs.push_back (Attribute (true, library->get_metadatum_id (DATUM_LOCATION), string(key)));
              library->del (attrs);
	      store_tracks->erase (*cur);

	      if (physically_delete_files)
	      {
		std::string filename = Glib::filename_from_uri (key);
		g_unlink (filename.c_str()); 	
	      }
          }

          if (store_tracks->children ().empty())
          {
              remove_album ();
          }
      }

      //////////////////////////// Play/Enqueue
      void
      LibraryUI::play_current (const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column)
      {
	  Gtk::TreeViewColumn *column_0 = view_albums->get_column (0);
	  if (column == column_0) return;

	  select_all ();
	  play ();
      }

      void
      LibraryUI::play ()
      {
	  Gtk::TreeSelection::ListHandle_Path paths = view_tracks->get_selection()->get_selected_rows ();
	  char **uri_list = (char**)calloc (paths.size()+1, sizeof(char*));
	  unsigned int counter = 0;
	  for (Gtk::TreeSelection::ListHandle_Path::const_iterator iter = paths.begin (); iter != paths.end (); ++iter)
	  {
	      Gtk::TreeModel::iterator cur = store_tracks->get_iter (*iter);
	      Glib::ustring str = (*cur)[tracks.uri];
	      uri_list[counter++] = strdup(str.c_str ());
	  }
	  uri_list[paths.size()] = 0;

	  int out;
	  bmp_system_control_add_uri_list (bmp_system_control,
					   (const char**)uri_list,
					   -1,
					   true, 
					   true, 
					   0, 
					   &out,
					   NULL);
	  g_strfreev (uri_list);
	  view_tracks->get_selection()->unselect_all ();
      }

      void
      LibraryUI::enqueue ()
      {
	  Gtk::TreeSelection::ListHandle_Path paths = view_tracks->get_selection()->get_selected_rows ();
	  char **uri_list = (char**)calloc (paths.size()+1, sizeof(char*));
	  unsigned int counter = 0;
	  for (Gtk::TreeSelection::ListHandle_Path::const_iterator iter = paths.begin (); iter != paths.end (); ++iter)
	  {
	      Gtk::TreeModel::iterator cur = store_tracks->get_iter (*iter);
	      Glib::ustring str = (*cur)[tracks.uri];
	      uri_list[counter++] = strdup(str.c_str ());
	  }
	  uri_list[paths.size()] = 0;

	  int out;
	  bmp_system_control_add_uri_list (bmp_system_control,
					   (const char**)uri_list,
					   -1,
					   false, 
					   false, 
					   -1, 
					   &out,
					   NULL);
	  g_strfreev (uri_list);
	  view_tracks->get_selection()->unselect_all ();
      }

      //////////////////////////// Adding Music/Rebuilding view 
      void
      LibraryUI::albums_merge ()
      {
	  using namespace Bmp::DB;
	  using namespace Bmp::Library;
	  using namespace Gtk;
	  using namespace Gdk;
	  using namespace Glib;
	  using namespace std;

	  Util::FileList list;
	  TreeModel::Children nodes = store_tracks->children ();
          for (TreeModel::Children::const_iterator node = nodes.begin (), end = nodes.end (); node != end; ++node)
          {
	    Glib::ustring uri = (*node)[tracks.uri];
	    list.push_back (uri);
	  }

	  Util::FileList non_writable;
	  Util::files_writable (list, non_writable);
	  if (!non_writable.empty ())
	  {
	    Gtk::MessageDialog dialog (*this, "Unable to merge: Some files are not writable.", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
	    dialog.run ();
	    return;
	  }

	  if (!Util::files_taggable (list))
	  {
	    Gtk::MessageDialog dialog (*this, "Unable to merge: Some files are currently not taggable with BMP.", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
	    dialog.run ();
	    return;
	  }

	  typedef TreeSelection::ListHandle_Path Paths;

	  MusicBrainzWS::MB_Releases mb_releases; 
	  Paths paths = view_albums->get_selection()->get_selected_rows ();
	  for (Paths::const_iterator path  = paths.begin (),
				     end   = paths.end ();
				     path != end; ++path)
	  {
	    TreeModel::iterator cur = store_albums_filter->convert_iter_to_child_iter (store_albums_filter->get_iter (*path)); 
	    ustring artist = (*cur)[albums.artist];
	    ustring album  = (*cur)[albums.album];
	    MusicBrainzWS::get_releases (artist, album, mb_releases);
	  }

	  LibraryUIMergeDialog *dialog = LibraryUIMergeDialog::create ();
	  if (dialog->run (mb_releases) != GTK_RESPONSE_OK)
	  {
	    delete dialog;
	    return;
	  }

	  MusicBrainzWS::MB_Release selected = dialog->selected;
	  delete dialog;

	  Glib::ustring title;
	  title.append (Markup::escape_text (selected.artist_name));
	  title.append ("\n");
	  title.append ("<i>");
	  title.append (Markup::escape_text (selected.title));
	  title.append ("</i>");

	  UpdateTracks u_tracks;
	  for (Util::FileList::const_iterator	file = list.begin (); file != list.end (); ++file)
	  {
	    UpdateTrack track;

	    track.location	    = (*file);
	    track.artist	    = selected.artist_name;
	    track.album		    = selected.title;
	    track.mb_album_artistid = selected.artist_id; 
	    track.mb_releaseid	    = selected.id; 

	    if (!selected.asin.empty()) track.asin = selected.asin;

	    u_tracks.push_back (track); 
	  }

	  busy = true;
	  library->modify_tracks (u_tracks);

	  typedef std::vector < Gtk::TreeModel::RowReference > VRowRefs;
	  VRowRefs vrefs;
	  for (Paths::const_iterator path  = paths.begin (),
				     end   = paths.end ();
				     path != end; ++path)
	  {
	    vrefs.push_back (Gtk::TreeModel::RowReference (store_albums_filter, (*path))); 
	  }

	  for (VRowRefs::const_iterator ref   = vrefs.begin (),
					end   = vrefs.end ();
				        ref  != end; ++ref)
	  {
	    TreeModel::RowReference const& r = (*ref); 
	    TreeModel::iterator filter = store_albums_filter->get_iter (r.get_path ());
	    TreeModel::iterator cur = store_albums_filter->convert_iter_to_child_iter (filter);
	    RowMap::iterator rowMapIter = row_map.find (RowKey((*cur)[albums.artist], (*cur)[albums.album]));
	    if (rowMapIter != row_map.end ())
	    {
		row_map.erase (rowMapIter);
	    }
	    store_albums->erase (cur);
	  }

	  // Before we insert, make sure we remove entries that possibly match our new entry, or we'll end up with dupes	
	  RowMap::iterator rowMapIter = row_map.find (RowKey(selected.artist_name, selected.title));
	  if (rowMapIter != row_map.end ())
	  {
	      Gtk::TreeModel::RowReference& ref = rowMapIter->second;
	      Gtk::TreeModel::iterator modelIter = store_albums->get_iter (ref.get_path ());
	      store_albums->erase (modelIter);
	      row_map.erase (rowMapIter);
	  }

	  TreeModel::iterator cur = store_albums->append ();
#ifdef USE_AMAZON
	  Glib::RefPtr<Gdk::Pixbuf> disc = 
		Gdk::Pixbuf::create_from_file (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "audio.png"))->scale_simple 
					  (64, 64, Gdk::INTERP_BILINEAR);
	  RefPtr<Pixbuf> cover;
	  try {
	    Amazon::get_cover_by_asin (selected.asin, cover, false);
	    (*cur)[albums.cover] = cover->scale_simple (64, 64, INTERP_BILINEAR);
	    (*cur)[albums.has_cover] = true;
	  }
	  catch (...)
	  {
	    (*cur)[albums.cover] = disc;
	    (*cur)[albums.has_cover] = false;
	  }

#else
	  (*cur)[albums.cover]  = disc;
#endif

	  (*cur)[albums.artist] = selected.artist_name;
	  (*cur)[albums.album]  = selected.title;
	  (*cur)[albums.title]  = title;

	  row_map.insert (std::make_pair (RowKey(selected.artist_name, selected.title), Gtk::TreeModel::RowReference (store_albums, Gtk::TreePath (cur))));
	  busy = false;

	  store_albums_filter->clear_cache ();
	  store_albums_filter->refilter ();
	  update_track_view_from_albums ();
      }

      void
      LibraryUI::music_add ()
      {
	Gtk::FileChooserDialog dialog (*this,
                                       _("Select Path To Add"),
                                       Gtk::FILE_CHOOSER_ACTION_SELECT_FOLDER);

	dialog.add_button (Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	dialog.add_button (Gtk::Stock::ADD, Gtk::RESPONSE_OK);
	dialog.set_current_folder (mcs->key_get<Glib::ustring>("bmp", "file-chooser-path"));

	if (dialog.run () == Gtk::RESPONSE_CANCEL)
	{
	   return;
	}
	else
	{
	  notebook_add->set_current_page (1);
	  Glib::ustring folder = dialog.get_filename ();
	  dialog.hide ();

	  busy = true; 

	  Util::FileList list;
	  Util::collect_path (folder, list);

	  for (Util::FileList::const_iterator iter = list.begin (), end = list.end ();
	       iter != end;
	       ++iter)
	  {
	    debug ("library-ui", G_STRLOC ": URI: '%s'", (*iter).c_str());
	    library->cache_metadata (*iter, folder);
	    while (gtk_events_pending ()) gtk_main_iteration ();
	  }

	  debug ("library-ui", "Updating compilations");
	  library->update_compilations ();
	  notebook_add->set_current_page (0);

	  debug ("library-ui", "Updating albums view");
	  albums_update ();

	  busy = false; 
	}
    }

    void
    LibraryUI::albums_refilter ()
    {
	store_albums_filter->refilter ();
    }

    bool
    LibraryUI::albums_row_is_visible (const Gtk::TreeModel::const_iterator& iter)
    {
	Glib::ustring artist ((*iter)[albums.artist]);	
	Glib::ustring album  ((*iter)[albums.album]);	
	Glib::ustring needle (albums_filter->get_text ());
  
	return (needle.empty() || match_keys_boost (artist + " " + album, needle));
    }

    void
    LibraryUI::albums_update ()
    {
	using namespace Glib; 
	using namespace Gtk; 
	using namespace Gdk; 
	using namespace Bmp::DB;
	using namespace Bmp::Library;
	using namespace std;
	using namespace boost;

	Glib::RefPtr<Gdk::Pixbuf> disc = 
		Gdk::Pixbuf::create_from_file (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "audio.png"))->scale_simple 
					  (64, 64, Gdk::INTERP_BILINEAR);

        string id_album    = library->get_metadatum_id (DATUM_ALBUM);
        string id_artist   = library->get_metadatum_id (DATUM_ARTIST);
	string id_location = library->get_metadatum_id (DATUM_LOCATION);

	store_albums->clear ();
	row_map.clear ();
   
	busy = true; 

        AttributeList attrs_albums;
        attrs_albums.push_back (Attribute (false, id_location, std::string("file://"))); //Make sure we only fetch files

        VDataRow vector_albums;
        library->project (id_album, vector_albums, attrs_albums);

        for (VDataRow::iterator iter = vector_albums.begin (), end = vector_albums.end (); iter != end; ++iter)
	{
          string album = get<Glib::ustring>((*iter)[id_album]);
          if (album.length())
          {
	      if (library->is_compilation (album)) continue;

	      AttributeList attrs_artists;
	      attrs_artists.push_back (Attribute (false, id_location, string("file://")));
	      attrs_artists.push_back (Attribute (true,  id_album, album));

	      VDataRow vector_artists;
	      library->project (id_artist, vector_artists, attrs_artists);

	      for (VDataRow::iterator iter = vector_artists.begin (), end = vector_artists.end (); iter != end; ++iter)
	      {
		string artist = get<Glib::ustring>((*iter)[id_artist]);

		if (artist.length())
		{
		  TreeModel::iterator cur = store_albums->append ();

		  Glib::ustring title;
		  title.append (Markup::escape_text (artist));
		  title.append ("\n");
		  title.append ("<i>"); 
		  title.append (Markup::escape_text (album));
		  title.append ("</i>");

#ifdef USE_AMAZON
		  RefPtr<Pixbuf> cover;
		  try {
		    Glib::ustring asin = library->get_asin_for_album (album);
		    Amazon::get_cover_by_asin (asin, cover, false);
		    (*cur)[albums.cover] = cover->scale_simple (64, 64, INTERP_BILINEAR);
		    (*cur)[albums.has_cover] = true;
		  }
		  catch (...)
		  {
		    (*cur)[albums.cover] = disc;
		    (*cur)[albums.has_cover] = false;
		  }

#else
		  (*cur)[albums.cover]  = disc;
#endif
		  (*cur)[albums.album]  = album;
		  (*cur)[albums.artist] = artist;
		  (*cur)[albums.title]  = title;

		  row_map.insert (std::make_pair (RowKey (artist, album), Gtk::TreeModel::RowReference (store_albums, Gtk::TreePath (cur))));

		  while (gtk_events_pending ()) gtk_main_iteration ();
		}
	      }
	  }
	}

	busy = false; 
    }

}
