#include <glibmm.h>
#include <glib/gi18n.h>

#include <ne_basic.h>
#include <ne_compress.h>
#include <ne_socket.h>
#include <ne_utils.h>
#include <ne_props.h>
#include <ne_session.h>
#include <ne_request.h>
#include <ne_uri.h>

#include <vector>

#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/replace.hpp>

#include <iostream>
#include <sstream>
#include <fstream>

#include "util.hpp"
#include "uri++.hpp"
#include "lyrics.hpp"
#include "xml.hpp"
#include "main.hpp"

#include "paths.hpp"

#include "x_vfs.hpp"

namespace
{
    const char* param[] = {
      "act",
      "cln",
      "clv",
      "ar",
      "ti",
      "al",
      "co",
      "yr",
      "txt",
      "and",
      "id",
    };

    enum ParamID
    {
      LT_ACTION,
      LT_CLIENT,
      LT_CLVERSION,
      LT_ARTIST,
      LT_TITLE,
      LT_ALBUM,
      LT_COMPOSER,
      LT_YEAR,
      LT_TEXT,
      LT_AND,
      LT_ID,
    };
}

namespace Bmp
{

  Lyrictracker::Lyrictracker ()
  {
  };

  Lyrictracker::~Lyrictracker ()
  {
  };

  bool 
  Lyrictracker::get_lyrics (const Bmp::Library::Track& track,
			    Glib::ustring&	       lyrics,
			    bool		       cached_only)
  {
      static const std::string LT_BASE_URI = "http://lyrictracker.com/soap.php";
      if (!(track.artist && track.title)) return false;

      Glib::ustring artist = track.artist.get ();  
      Glib::ustring title  = track.title.get ();
      std::string basename = (boost::format ("%1%_%2%.txt") % artist.lowercase () % title.lowercase ()).str ();
      std::string text_path = Glib::build_filename (BMP_PATH_LYRIC_CACHE_DIR, basename);

      if (Glib::file_test (text_path, Glib::FILE_TEST_EXISTS))
      {
	  lyrics = Glib::file_get_contents (text_path);
	  return true;
      }
  
      if (cached_only) return false;
    
      std::stringstream uri_query_id;
      uri_query_id << LT_BASE_URI << "?"
	  << param[LT_CLIENT] << "=bmp&"
	  << param[LT_CLVERSION] << "=0.20&"
	  << param[LT_TITLE] << "=" << track.title.get () << "&"
	  << param[LT_ARTIST] << "=" << track.artist.get () << "&"
	  << param[LT_ACTION] << "=query&"
	  << param[LT_AND] << "=1";

      std::string auxilliary = uri_query_id.str();
      boost::replace_all (auxilliary, "'", "\\'");
	 
      Bmp::URI u_query_id (auxilliary); 
      u_query_id.escape ();
      Glib::ustring uri_query_id_utf8 = Glib::ustring(u_query_id);
      Bmp::VFS::Handle handle_id = Bmp::VFS::Handle::create (uri_query_id_utf8);
      vfs->read (handle_id, Bmp::VFS::TRANSPORT);
      if (!handle_id.get_buffer()) return false;

      xmlDocPtr doc = 0;
      xmlXPathObjectPtr xpathobj = 0;
      xmlNodeSetPtr nv = 0;

      doc = xmlParseMemory (reinterpret_cast<const char*>(handle_id.get_buffer()), handle_id.get_buffer_size());
      if (!doc) return false;

      xpathobj = xml_execute_xpath_expression (doc, BAD_CAST "//result", NULL);
      nv = xpathobj->nodesetval;
      if ((!nv) || (!nv->nodeNr)) return false;
      xmlChar *id;
      id = xmlGetProp (nv->nodeTab[0], BAD_CAST "id");
      if (!id) return false;

      std::stringstream uri_query_lyrics;
      uri_query_lyrics << LT_BASE_URI << "?"
	  << param[LT_CLIENT] << "=bmp&"
	  << param[LT_CLVERSION] << "=0.20&"
	  << param[LT_ID] << "=" << id << "&"
	  << param[LT_ACTION] << "=detail";

      free (id);
    
      Bmp::URI u_query_lyrics (uri_query_lyrics.str()); 
      u_query_lyrics.escape ();
      Glib::ustring uri_query_lyrics_utf8 = Glib::ustring(u_query_lyrics);
      Bmp::VFS::Handle handle_lyrics = Bmp::VFS::Handle::create (uri_query_lyrics_utf8);
      vfs->read (handle_lyrics, Bmp::VFS::TRANSPORT);
      if (!handle_lyrics.get_buffer()) return false;

      lyrics = reinterpret_cast<const char*>(handle_lyrics.get_buffer ());

      try {
	std::ofstream outfile;      
	outfile.open (text_path.c_str ());
	outfile << std::string(lyrics); 
	outfile.close ();
      } catch (Glib::ConvertError& cxe) { g_warning ("Unable to save lyrics for '%s' by '%s'", title.c_str(), artist.c_str()); }

      return true;
  }
}
