//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "paths.hpp"

#include <gtkmm.h>
#include <dbus/dbus-glib.h>

#include "notify.hpp"
#include "debug.hpp"

#include <libnotify/notify.h>

namespace Bmp
{
    Notify::Notify () : visible (false)
    {
      // Init libnotify
      notify_init (PACKAGE);

      // Make sure notify-daemon is running
      DBusGConnection *bus;
      DBusGProxy      *obj;
      GError	      *error = NULL;
      char	      *introspect_data;

      bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
      if (!bus)
      {
	debug ("notify", "Unable to get session bus");
	throw UNABLE_TO_DBUS_CONNECT;
      }

      obj = dbus_g_proxy_new_for_name (bus,
                                       "org.freedesktop.Notifications", 
                                       "/org/freedesktop/Notifications", 
                                       DBUS_INTERFACE_INTROSPECTABLE);

      if (!dbus_g_proxy_call (obj, "Introspect", &error,
                          G_TYPE_INVALID,
                          G_TYPE_STRING, &introspect_data, G_TYPE_INVALID))
      {
	debug ("notify", "Unable to introspect");
	throw UNABLE_TO_INTROSPECT_INTERFACE;
      }
      free (introspect_data);

      dbus_g_connection_unref (bus);     
      g_object_unref (obj);

      // Load default icon
      Glib::ustring filename = Glib::build_filename (Glib::build_filename 
						  (Glib::build_filename (DATA_DIR, BMP_BASENAME_ICON_DIR), "red"),
						  "icon_player_32.png");
      default_icon = Gdk::Pixbuf::create_from_file (filename);

      n = notify_notification_new ("(none)",
				   "(none)", 
				   "(none)",
				   NULL); 
      notify_notification_set_icon_from_pixbuf (n, default_icon->gobj());
      g_signal_connect (G_OBJECT (n), "closed", G_CALLBACK (Notify::notification_closed), this);
    }

    Notify::~Notify ()
    {
      g_object_unref (n);
      notify_uninit ();
    }

    void
    Notify::notification_closed (NotifyNotification *n, Bmp::Notify *notify)
    {
      notify->visible = false;	
      g_object_unref (n);

      notify->n = notify_notification_new ("(none)",
				   "(none)", 
				   "(none)",
				   NULL); 
      notify_notification_set_icon_from_pixbuf (notify->n, notify->default_icon->gobj());
      g_signal_connect (G_OBJECT (notify->n), "closed", G_CALLBACK (Notify::notification_closed), notify);
    }

    void
    Notify::set_notification (const Glib::ustring& summary,
			      const Glib::ustring& message,
			      unsigned int	 timeout) // We multiply it by 1k so we can just pass seconds here
    {
      GError *error = 0;
      notify_notification_update (n, (!summary.empty() ? summary.c_str () : ""),
				  (!message.empty() ? message.c_str () : ""),
				  NULL);
      notify_notification_set_timeout (n, timeout * 1000);
      notify_notification_set_icon_from_pixbuf (n, default_icon->gobj());
      notify_notification_show (n, &error);
      visible = true;
    }

    void
    Notify::set_notification (const Glib::ustring& summary,
			      const Glib::ustring& message,
			      unsigned int	 timeout, // We multiply it by 1k so we can just pass seconds here
			      Glib::RefPtr<Gdk::Pixbuf>& pixbuf)
    {
      GError *error = 0;
      notify_notification_update (n, (!summary.empty() ? summary.c_str () : ""),
				  (!message.empty() ? message.c_str () : ""),
				  NULL);
      notify_notification_set_timeout (n, timeout * 1000);
      notify_notification_set_icon_from_pixbuf (n, pixbuf->scale_simple (80, 80, Gdk::INTERP_BILINEAR)->gobj());
      notify_notification_show (n, &error);
      visible = true;
    }

} // ns::Bmp
