//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non GPL-compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_GUI
#  include <gtk/gtk.h>
#endif

#include <glibmm.h>
#include <glib/gstdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include <goa/libgoa.h>

#include "database.hpp"
#include "file_utils.hpp"
#include "interfaces.h"
#include "library.hpp"
#include "loader.hpp"
#include "playlist.hpp"
#include "plugin.h"
#include "uri++.hpp"
#include "util.hpp"
#include "vfs.hpp"

#include "main.hpp"
#include "paths.hpp"

#include "bmp-marshalers.h"

#include "x_library.hpp"
#include "x_play.hpp"
#include "x_playlist.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

#include "gmacros.hpp"
BMP_DEFINE_TYPE (BmpPlaylist, bmp_playlist, G_TYPE_OBJECT)

enum BmpPlaylistSignals
{
        BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED,
        BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED,

        BMP_PLAYLIST_N_SIGNALS
};

namespace { guint signals[BMP_PLAYLIST_N_SIGNALS] = { 0, 0 }; }

struct _BmpPlaylistPrivate
{
        bool dispose_has_run;
        GtkListStore *playlist;
	GHashTable   *row_guid_map;
	GHashTable   *uri_iter_map;
};

namespace
{

  Bmp::Library::Datum vcolumns[] =
  {
    Bmp::Library::DATUM_ARTIST,
    Bmp::Library::DATUM_ALBUM,
    Bmp::Library::DATUM_TRACK,
    Bmp::Library::DATUM_TITLE,
    Bmp::Library::DATUM_GENRE,
    Bmp::Library::DATUM_DATE,
    Bmp::Library::DATUM_TIME
  };

  void
  load_playlist (BmpPlaylist *tracklist)
  {
        gchar    **uri_strv;
        int       new_count = 0;

        uri_strv = g_new0 (gchar*,2);
        uri_strv[0] = g_strconcat ("file://", BMP_PATH_USER_DIR, "/tracklist.xspf", NULL);
        uri_strv[1] = NULL;

        bmp_system_control_add_uri_list (bmp_system_control,
                                        (const gchar**)uri_strv,
                                        -1,
                                        false,
                                        false,
                                        0,
                                        &new_count,
                                        NULL);
        g_strfreev (uri_strv);
  }

  void
  save_playlist (BmpPlaylist *tracklist)
  {
	Bmp::Util::FileList list;
        GtkTreeIter	    iter;
        bool		    proceed;

        proceed = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(tracklist->priv->playlist), &iter);
        while (proceed)
        {
            char *uri;
            gtk_tree_model_get (GTK_TREE_MODEL(tracklist->priv->playlist), &iter, COLUMN_URI, &uri, -1);
	    Bmp::URI u (uri);
	    free (uri);
	    u.escape ();
	    list.push_back (Glib::ustring(u));
            proceed = gtk_tree_model_iter_next (GTK_TREE_MODEL(tracklist->priv->playlist), &iter);
        }

	Glib::ustring filename = Glib::filename_to_uri (Glib::build_filename (BMP_PATH_USER_DIR, "tracklist.xspf"));

	try
	{
	  vfs->write (list, filename, "xspf");
	}
	catch (Bmp::VFS::Exception& cxe)
	{
	  g_warning ("Couldn't save playlist to %s", filename.c_str ());
	}
  }

  void
  update_item  (BmpPlaylist		    *tracklist,
		GtkTreeIter		    *iter,
		Bmp::Library::UpdateTrack&   track)
  {
      using namespace Bmp::DB;
      using namespace Bmp::Library;

      GtkListStore *store = tracklist->priv->playlist;

      if (track.artist)
	gtk_list_store_set (store, iter, 0, track.artist.get().c_str(), -1);

      if (track.album)
	gtk_list_store_set (store, iter, 1, track.album.get().c_str(), -1);

      if (track.tracknumber)
	gtk_list_store_set (store, iter, 2, track.tracknumber.get(), -1);

      if (track.title)
	gtk_list_store_set (store, iter, 3, track.title.get().c_str(), -1);

      if (track.genre)
	gtk_list_store_set (store, iter, 4, track.genre.get().c_str(), -1);

      if (track.date)
	gtk_list_store_set (store, iter, 5, track.date.get(), -1);

  }

  void
  playlist_update_item (Glib::ustring		     pkey,
			Bmp::Library::UpdateTrack    track,
			BmpPlaylist		    *tracklist)
  {
    g_return_if_fail (BMP_IS_PLAYLIST (tracklist));

    gpointer p = g_hash_table_lookup (tracklist->priv->uri_iter_map, pkey.c_str());
    if (!p) return;
    GtkTreeIter *iter = reinterpret_cast<GtkTreeIter*>(p);
    update_item (tracklist, iter, track);
  }

  int
  insert_items (BmpPlaylist	  *tracklist,
                const char	  *uri,
                int                index)
  {
    using namespace Bmp::DB;

    Bmp::URI u (uri);
    Bmp::URI::Protocol p = u.get_protocol ();
    u.unescape ();

    if ((!vfs->has_container (uri) || ((play->is_audio_file (uri) || (p == Bmp::URI::PROTOCOL_CDDA)))) && !Glib::file_test (u.path, Glib::FILE_TEST_IS_DIR))
      {
        DataRow row;
	try {
	  library->get_metadata (Glib::ustring(uri), row);
	} catch (Bmp::Library::Exception& cxe) {
	  return index; 
	}

	DataRow::iterator end = row.end ();
        GtkTreeIter iter;
        gtk_list_store_insert (GTK_LIST_STORE(tracklist->priv->playlist), &iter, index);

        for (unsigned int n = 0; n < G_N_ELEMENTS(vcolumns); n++)
        {
	  DataRow::iterator row_iter = row.find(library->get_metadatum_id (vcolumns[n]));

	  if (row_iter != end) 
	  {
	    if (row_iter->second.which() == VALUE_TYPE_STRING)
	    {
                gtk_list_store_set (GTK_LIST_STORE(tracklist->priv->playlist), &iter,
                                    n, (boost::get<Glib::ustring>(row_iter->second).c_str()), -1);
	    }
	    else if (row_iter->second.which() == VALUE_TYPE_INT)
	    {
                gtk_list_store_set (GTK_LIST_STORE(tracklist->priv->playlist), &iter,
                                    n, boost::get<int>(row_iter->second), -1);
	    }
	  }
	}
	RowGUID *guid = bmp_row_guid_new (uri);
	gtk_list_store_set (GTK_LIST_STORE(tracklist->priv->playlist), &iter,
                            COLUMN_GUID,   guid,
                            COLUMN_URI,    uri, 
                            -1);

	//Insert GUID into mapping
	g_hash_table_insert (tracklist->priv->row_guid_map, guid, g_boxed_copy (GTK_TYPE_TREE_ITER, &iter));
	g_hash_table_insert (tracklist->priv->uri_iter_map, g_strdup (uri), g_boxed_copy (GTK_TYPE_TREE_ITER, &iter));
	index++;
      }
    else
      {
	Bmp::Util::FileList list;
  
	if ((u.get_protocol() == Bmp::URI::PROTOCOL_FILE) && (Glib::file_test (u.path, Glib::FILE_TEST_IS_DIR)))
	{
	  Bmp::Util::collect_path (u.path, list);
	}
	else
	{
	  try
	  {
	    Bmp::VFS::Handle handle = Bmp::VFS::Handle::create (Glib::ustring(uri));
	    if (u.get_protocol() == Bmp::URI::PROTOCOL_QUERY)
	    {
	      vfs->read (handle, list, Bmp::VFS::CONTAINER);
	    }
	    else
	    {
	      vfs->read (handle, list);
	    }
	  }
	  catch (...) { return index; }
	}

	for (Bmp::Util::FileList::const_iterator i = list.begin ();
	    i != list.end (); 
	    ++i)
	{
		Bmp::URI uri (*i);
		DataRow row;
		DataRow::iterator end = row.end ();

		try { library->get_metadata (*i, row); } catch (Bmp::Library::Exception& cxe) { continue; }

		GtkTreeIter iter;
		gtk_list_store_insert (GTK_LIST_STORE(tracklist->priv->playlist), &iter, index);

		for (unsigned int n = 0; n < G_N_ELEMENTS(vcolumns); n++)
		{
		  DataRow::iterator row_iter = row.find(library->get_metadatum_id (vcolumns[n]));
		  if (row_iter != end)
		  {
		    if (row_iter->second.which() == VALUE_TYPE_STRING)
		    {
		      gtk_list_store_set (GTK_LIST_STORE(tracklist->priv->playlist), &iter,
                                    n, (boost::get<Glib::ustring>(row_iter->second).c_str()), -1);
		    }
		    else if (row_iter->second.which() == VALUE_TYPE_INT)
		    {
		      gtk_list_store_set (GTK_LIST_STORE(tracklist->priv->playlist), &iter,
                                    n, boost::get<int>(row_iter->second), -1);
		    }
		  }
		}

		RowGUID *guid = bmp_row_guid_new (Glib::ustring(uri).c_str());
		DataRow::iterator row_iter = row.find(library->get_metadatum_id (Bmp::Library::DATUM_LOCATION));
		std::string real_uri = boost::get<Glib::ustring>(row_iter->second);

		gtk_list_store_set (GTK_LIST_STORE(tracklist->priv->playlist), &iter,
				      COLUMN_GUID,   guid,
				      COLUMN_URI,    real_uri.c_str(), 
				      -1);

		//Insert GUID into mapping
		g_hash_table_insert (tracklist->priv->row_guid_map, guid, g_boxed_copy (GTK_TYPE_TREE_ITER, &iter));
		g_hash_table_insert (tracklist->priv->uri_iter_map, g_strdup (real_uri.c_str()), g_boxed_copy (GTK_TYPE_TREE_ITER, &iter));
		index++;
            }
      }
      return index;
  }

  void 
  free_tree_iter (gpointer iter)
  {
    g_boxed_free (GTK_TYPE_TREE_ITER, iter); 
  }

  //GObject
  void
  bmp_playlist_init (BmpPlaylist  *tracklist)
  {
        tracklist->priv = g_new0(BmpPlaylistPrivate, 1);

        tracklist->priv->dispose_has_run = false;
        tracklist->priv->playlist = gtk_list_store_new (9,
                                                   G_TYPE_STRING,
                                                   G_TYPE_STRING,
                                                   G_TYPE_INT,
                                                   G_TYPE_STRING,
                                                   G_TYPE_STRING,
                                                   G_TYPE_INT,
                                                   G_TYPE_INT,
                                                   BMP_TYPE_ROW_GUID, //GUID
                                                   G_TYPE_STRING); //URI

  
	tracklist->priv->row_guid_map = g_hash_table_new_full (GHashFunc(row_guid_hash_func),
							  GEqualFunc(row_guid_equal_func),
							  GDestroyNotify(bmp_row_guid_free),
							  GDestroyNotify(free_tree_iter));

	tracklist->priv->uri_iter_map  = g_hash_table_new_full (g_str_hash,
							  g_str_equal, 
							  GDestroyNotify(g_free),
							  GDestroyNotify(free_tree_iter));

        g_object_set_data (G_OBJECT(tracklist->priv->playlist), "current", NULL);
  }

  GObject*
  bmp_playlist_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
  {
        GObject		  *obj;
	BmpPlaylistClass  *klass;
	GObjectClass	  *parent_class;

	klass	      = BMP_PLAYLIST_CLASS (g_type_class_peek (BMP_TYPE_PLAYLIST));
	parent_class  = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj	      = parent_class->constructor (type,
                                                   n_construct_properties,
                                                   construct_properties);
        return obj;
  }

  void
  bmp_playlist_dispose (GObject *obj)
  {
    BmpPlaylist *tracklist = (BmpPlaylist*) obj;

    if (tracklist->priv->dispose_has_run) return;
    tracklist->priv->dispose_has_run = true;

    save_playlist (tracklist);
  }

  void
  bmp_playlist_finalize (GObject *object)
  {
    BmpPlaylist *tracklist = (BmpPlaylist*)object;
    g_free (tracklist->priv);
  }

  void
  bmp_playlist_class_init (BmpPlaylistClass *g_class)

  {
    GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

    gobject_class->dispose      = bmp_playlist_dispose;
    gobject_class->finalize     = bmp_playlist_finalize;
    gobject_class->constructor  = bmp_playlist_constructor;

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED] =
        g_signal_new ("tracklist-items-removed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      GSignalFlags (0),
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_items_removed),
                      NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                      2, G_TYPE_POINTER, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED] =
        g_signal_new ("tracklist-items-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      GSignalFlags (0),
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_items_added),
                      NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                      2, G_TYPE_POINTER, G_TYPE_INT);
  }

}

unsigned int
vcolumns_get_n_items (void)
{
    return G_N_ELEMENTS (vcolumns);
}

Bmp::Library::Datum
vcolumns_get_item_nth (unsigned int n)
{
  return vcolumns[n];
}

BmpPlaylist*
bmp_playlist_new (void)
{
  BmpPlaylist *tracklist;

  tracklist = static_cast<BmpPlaylist *> (g_object_new (bmp_playlist_get_type (), NULL));

  return tracklist;
}

GtkListStore*
bmp_playlist_get_playlist (BmpPlaylist *tracklist)
{
  return tracklist->priv->playlist;
}

void
bmp_playlist_row_removing (BmpPlaylist *tracklist, GtkTreePath *path)
{
    g_return_if_fail (BMP_IS_PLAYLIST (tracklist));

    RowGUID *guid = 0;
    char *uri = 0;
    GtkTreeIter iter;  

    if (!gtk_tree_model_get_iter (GTK_TREE_MODEL (tracklist->priv->playlist), &iter, path)) return;
    gtk_tree_model_get (GTK_TREE_MODEL (tracklist->priv->playlist), &iter, COLUMN_GUID, &guid, COLUMN_URI, &uri, -1);
    if (!guid) return;
    bmp_system_control_remove_guid (bmp_system_control, guid);
    g_hash_table_remove (tracklist->priv->row_guid_map, guid);
    g_hash_table_remove (tracklist->priv->uri_iter_map, uri);
    g_free (uri);
    bmp_row_guid_free (guid);
}

int
bmp_playlist_get_row_by_guid (BmpPlaylist *tracklist, RowGUID *guid)
{
    g_return_val_if_fail (BMP_IS_PLAYLIST (tracklist), -1);

    gpointer p = g_hash_table_lookup (tracklist->priv->row_guid_map, guid);
    if (!p) return -1;
    GtkTreeIter *iter = reinterpret_cast<GtkTreeIter*>(p);
    GtkTreePath *path = gtk_tree_model_get_path (GTK_TREE_MODEL(tracklist->priv->playlist), iter);
    if (!path) return -1;
    int index = gtk_tree_path_get_indices (path)[0];
    gtk_tree_path_free (path);
    return index;
}

bool
bmp_playlist_get_row_by_index (BmpPlaylist        *tracklist,
                               int                 index,
                               GtkTreeIter        *iter)
{
  return gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(tracklist->priv->playlist), iter, NULL, index);
}

char*
bmp_playlist_get_uri_by_index (BmpPlaylist        *tracklist,
                               int                 index)

{
    GtkTreeIter  iter;
    char        *uri;

    if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(tracklist->priv->playlist), &iter, NULL, index)) return NULL;
    gtk_tree_model_get (GTK_TREE_MODEL(tracklist->priv->playlist), &iter, COLUMN_URI, &uri, -1);
    return uri;
}

int
bmp_playlist_insert_tracklist_items_p (BmpPlaylist *tracklist, const char **uri_list, int idx)
{
    int	    n_added_rows = 0,
	    idx_original = idx;
    GArray *added_rows;
  
    int n = 0;
    while (uri_list[n])
      {
        idx = insert_items (tracklist, uri_list[n], idx);
        n++;
      }

    added_rows = g_array_new (false, false, sizeof(int));
    n_added_rows = (idx - idx_original);

    for (int n = 0; n < n_added_rows; ++n)
    {
        int val = n + idx_original;
        g_array_append_val (added_rows, val);
    }

    g_signal_emit (G_OBJECT (tracklist),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED],
                   0, added_rows, n_added_rows);

    return idx;
}

int
bmp_playlist_insert_tracklist_items (BmpPlaylist *tracklist, const gchar **uri_list)
{
   return bmp_playlist_insert_tracklist_items_p (tracklist, uri_list, gtk_tree_model_iter_n_children (GTK_TREE_MODEL(tracklist->priv->playlist), NULL));
}

void
bmp_playlist_tracklist_remove_rows_all (BmpPlaylist *tracklist)
{
    GArray                  *removed_rows;
    int                     n_removed_rows = -1;

    gtk_tree_row_reference_free (static_cast<GtkTreeRowReference *> (g_object_get_data (G_OBJECT(tracklist->priv->playlist), "current")));
    g_object_set_data (G_OBJECT(tracklist->priv->playlist), "current", NULL);
    gtk_list_store_clear (tracklist->priv->playlist);

    removed_rows = g_array_new (false,false,sizeof(int));

    g_signal_emit (G_OBJECT (tracklist),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, removed_rows, n_removed_rows);
}

void
bmp_playlist_tracklist_remove_selected (BmpPlaylist *tracklist, GList *paths)
{
    GArray *removed_rows;
    int    n_removed_rows;

    removed_rows = g_array_new (false, false, sizeof(int));
    n_removed_rows = g_list_length (paths);

    for ( ; paths ; paths = paths->next)
      {
	  GtkTreeIter  iter;
	  GtkTreePath *path = static_cast<GtkTreePath *> (paths->data);
	  int         row = gtk_tree_path_get_indices (path)[0];

	  if (gtk_tree_model_get_iter (GTK_TREE_MODEL(tracklist->priv->playlist), &iter, path))
	    {
	      gtk_list_store_remove (GTK_LIST_STORE(tracklist->priv->playlist), &iter);
	      g_array_append_val (removed_rows, row);
	    }
      }

    g_signal_emit (G_OBJECT (tracklist),
	       signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
	       0, removed_rows, n_removed_rows);
}

int
bmp_playlist_get_length (BmpPlaylist *tracklist)
{
    return gtk_tree_model_iter_n_children (GTK_TREE_MODEL(tracklist->priv->playlist), NULL);
}

char*
bmp_playlist_get_uri_nth (BmpPlaylist       *tracklist,
                          int                index)
{
    GtkTreeIter  iter;
    char        *uri;

    gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(tracklist->priv->playlist), &iter, NULL, index);
    gtk_tree_model_get (GTK_TREE_MODEL(tracklist->priv->playlist), &iter, COLUMN_URI, &uri, -1);
    return uri;
}

char*
bmp_playlist_get_current_uri (BmpPlaylist *tracklist)
{
    GtkTreeRowReference     *reference;
    GtkTreePath             *path;
    GtkTreeIter              iter;
    char                    *uri;

    reference = static_cast<GtkTreeRowReference *> (g_object_get_data (G_OBJECT(tracklist->priv->playlist), "current"));
    if (!reference) return NULL;

    path = gtk_tree_row_reference_get_path (reference);
    gtk_tree_model_get_iter (GTK_TREE_MODEL(tracklist->priv->playlist), &iter, path);
    gtk_tree_path_free (path);
    gtk_tree_model_get (GTK_TREE_MODEL(tracklist->priv->playlist), &iter, COLUMN_URI, &uri, -1);
    return uri;
}

char*
bmp_playlist_get_title_nth (BmpPlaylist         *tracklist,
                            int                  index)
{
    GtkTreeIter              iter;
    char                    *artist, *album, *title;

    gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(tracklist->priv->playlist), &iter, NULL, index);
    gtk_tree_model_get (GTK_TREE_MODEL(tracklist->priv->playlist),
			&iter,
			0, &artist,
			1, &album,
			3, &title, -1);

    std::stringstream titlestring;
    titlestring << (artist ? artist: "") << " - " << (album ? album:"") << " - " << (title ? title:""); 
    free (artist);
    free (album);
    free (title);
    return strdup (titlestring.str().c_str());
}

char*
bmp_playlist_get_current_title (BmpPlaylist *tracklist)
{
    GtkTreeRowReference     *reference;
    GtkTreePath             *path;
    GtkTreeIter              iter;
    char                    *artist, *album, *title;

    reference = static_cast<GtkTreeRowReference *> (g_object_get_data (G_OBJECT(tracklist->priv->playlist), "current"));
    if (!reference) return NULL;

    path = gtk_tree_row_reference_get_path (reference);
    if (!path) return NULL;

    gtk_tree_model_get_iter (GTK_TREE_MODEL(tracklist->priv->playlist), &iter, path);
    gtk_tree_path_free (path);

    gtk_tree_model_get (GTK_TREE_MODEL(tracklist->priv->playlist),
			&iter,
			0, &artist,
			1, &album,
			3, &title, -1);

    std::stringstream titlestring;
    titlestring << (artist ? artist: "") << " - " << (album ? album:"") << " - " << (title ? title:""); 
    free (artist);
    free (album);
    free (title);
    return strdup (titlestring.str().c_str());

}

void
bmp_playlist_load_playlist (BmpPlaylist *tracklist)
{
    g_return_if_fail (tracklist != NULL);
    g_return_if_fail (BMP_IS_PLAYLIST(tracklist));

    load_playlist (tracklist);
}

void
bmp_playlist_connect_library		    (BmpPlaylist *playlist)
{
    library->signal_row_updated().connect (sigc::bind (sigc::ptr_fun (&playlist_update_item), playlist));
}
