/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */


#ifndef BMP_WINAMP2_SKIN_HPP
#define BMP_WINAMP2_SKIN_HPP

#include <gtk/gtk.h>
#include <glib-object.h>

#define BMP_TYPE_WINAMP2_SKIN		  (bmp_winamp2_skin_get_type ())
#define BMP_WINAMP2_SKIN(obj)		  (G_TYPE_CHECK_INSTANCE_CAST ((obj), BMP_TYPE_WINAMP2_SKIN, BmpWinamp2Skin))
#define BMP_WINAMP2_SKIN_CLASS(klass)	  (G_TYPE_CHECK_CLASS_CAST ((klass),  BMP_TYPE_WINAMP2_SKIN, BmpWinamp2SkinClass))
#define BMP_IS_WINAMP2_SKIN(obj)	  	  (G_TYPE_CHECK_INSTANCE_TYPE ((obj), BMP_TYPE_WINAMP2_SKIN))
#define BMP_IS_WINAMP2_SKIN_CLASS(klass)    (G_TYPE_CHECK_CLASS_TYPE ((klass),  BMP_TYPE_WINAMP2_SKIN))
#define BMP_WINAMP2_SKIN_GET_CLASS(obj)     (G_TYPE_INSTANCE_GET_CLASS ((obj),  BMP_TYPE_WINAMP2_SKIN, BmpWinamp2SkinClass))

struct BmpWinamp2SkinClass
{
    GObjectClass parent;
};

struct BmpWinamp2Skin
{
    GObject parent;
};

#define SET_CURSOR_FOR_WIDGET(cursor_enum,widget) \
  { \
    GdkCursor *cursor; \
      if (!GTK_WIDGET_REALIZED(widget)) gtk_widget_realize(GTK_WIDGET(widget)); \
      cursor = bmp_cursor_get (CursorId(cursor_enum)); \
      gdk_window_set_cursor (GTK_WIDGET(widget)->window, cursor); \
      if (cursor) \
	gdk_cursor_destroy (cursor); \
  }

#define BMP_DEFAULT_FONT		"Sans 10"
#define BMP_DEFAULT_SKIN		"default"

// Non-action defines
#define BMP_FILE_REGION			    "region.txt"
#define BMP_FILE_PLEDIT			    "pledit.txt"
#define BMP_MASK_REGION_SECTION_MAIN_WINDOW "normal"
#define BMP_MASK_REGION_SECTION_EQUALIZER   "equalizer"

enum BmpSkinComponent
{
    SKIN_COMPONENT_MAIN,
    SKIN_COMPONENT_SHUFREP,
    SKIN_COMPONENT_CBUTTONS,
    SKIN_COMPONENT_POSBAR,
    SKIN_COMPONENT_VOLUME,
    SKIN_COMPONENT_TITLEBAR,
    SKIN_COMPONENT_BALANCE,
    SKIN_COMPONENT_PLEDIT,
    SKIN_COMPONENT_MONOSTER,
    SKIN_COMPONENT_NUMBERS,
    SKIN_COMPONENT_PLAYPAUSE,

    N_SKIN_COMPONENTS
};

enum BmpSkinColor
{
    BMP_SKINCOLOR_PL_BG_NORMAL,
    BMP_SKINCOLOR_PL_FG_NORMAL,
    BMP_SKINCOLOR_PL_BG_SELECTED,
    BMP_SKINCOLOR_PL_FG_CURRENT,
    BMP_N_SKINCOLORS
};

enum CursorId
{
    CURSOR_NORMAL,
    CURSOR_POSBAR,
    CURSOR_VOLBAR,
    CURSOR_PLAYLIST,
    CURSOR_PLAYLIST_SIZE,
    CURSOR_MAINMENU,
    CURSOR_N_CURSORS
};

enum SkinColorId
{
    SKIN_PLEDIT_NORMAL = 0,
    SKIN_PLEDIT_CURRENT,
    SKIN_PLEDIT_NORMALBG,
    SKIN_PLEDIT_SELECTEDBG,
    SKIN_TEXTBG,
    SKIN_TEXTFG,
    SKIN_N_COLORS
};

struct BmpTooltipEntry
{
    char *widget_name;
    char *tooltip_text;
    char *stock_id;
};

struct BmpButtonNormalT
{
    char *action_id;
    char *tooltip;
    int pixmap_id;             /* id of used pixmap */
    int img_x;
    int img_y;
    int img_x_pressed;
    int img_y_pressed;
    int width;
    int height;
    int pos_x;
    int pos_y;
};

struct BmpButtonToggleT
{
    char *action_id;
    char *tooltip;
    int pixmap_id;             /* id of used pixmap */
    int img_x;
    int img_y;
    int img_x_pressed;
    int img_y_pressed;
    int img_x_active;
    int img_y_active;
    int width;                 /* width in .bmp file */
    int height;                /* height in .bmp file */
    int pos_x;                 /* x position in skin when displayed */
    int pos_y;                 /* y position in skin when displayed */
    gboolean clickable;         /* is it clickable? */
};

struct BmpSkinComponentT
{
    BmpSkinComponent   identifier;
    char	      *filename;
    GdkPixbuf	      *pixbuf;
    gpointer	       user_data;
};

struct BmpWindowT
{
    GtkWidget *window;
};

struct BmpPosbarT
{
    BmpSkinComponent  identifier;
    char	      *filename;
    GdkPixbuf	      *pixbuf;
    int		      x_slider;
    gboolean	      dragging;
};

struct BmpVolumeT
{
    BmpSkinComponent  identifier;
    char	      *filename;
    GdkPixbuf	      *pixbuf;
    int		      x_slider;
    int		      height;
    gboolean	      dragging;
    gboolean	      has_button;
};

#define BMP_WINDOW_MAIN_WIDTH 275
#define BMP_WINDOW_MAIN_HEIGHT 116

enum BmpSkinType
{
    BMP_SKIN_ARCHIVE,
    BMP_SKIN_DIR
};

enum BmpWinamp2Window
{
    BMP_WINDOW_MAIN,
    BMP_WINDOW_PLEDIT,
};

extern BmpPosbarT	 *bmp_posbar;
extern BmpVolumeT	 *bmp_volume;
#if 0
extern BmpVolumeT	 *bmp_balance;
#endif
extern BmpSkinComponentT *bmp_skin_component[N_SKIN_COMPONENTS];
extern GdkBitmap	 *mask_window_main;

GType		    bmp_winamp2_skin_get_type	    (void);
BmpWinamp2Skin*	    bmp_winamp2_skin_new	    (void);
bool		    bmp_winamp2_skin_load_skin	    (BmpWinamp2Skin* skin, const std::string& path);

// Utility functions for main/playlist
GdkPixbuf*
bmp_pixbuf_new_simple (guint width, guint height);

void
bmp_pixbuf_copy_area_simple (GdkPixbuf *dest, guint x, guint y, guint width, guint height);

void
bmp_draw_pixbuf_simple (cairo_t *cr, GdkPixbuf *buf, guint x, guint y);

void
bmp_cursors_init ();

void
bmp_cursors_free ();

void
bmp_cursors_set (const std::string& archive_dirname);

void
bmp_cursor_add  (const std::string& archive_dirname,
	         int		    cursor_enum);

GdkCursor*
bmp_cursor_get  (::CursorId cursor_enum);

std::string
bmp_skin_get_color (const std::string&	path,
		    const std::string&  key);

GdkColor*
bmp_skin_convert_color (const std::string& value_arg);

void
bmp_colors_set_color (BmpSkinColor color, const std::string& color_hex);

void
bmp_colors_init (const std::string& archive_dirname);

void
bmp_colors_free ();

void
bmp_color_allocate (BmpSkinColor color, GtkWidget *widget);

GdkColor*
bmp_colors_get_color (BmpSkinColor color);

void
gdk_color_to_rgba (GdkColor *color, double *r, double *g, double *b, double *a);

#endif // BMP_WINAMP2_SKIN_HPP
